@@ -1,7 +1,7 @@
 #!/usr/bin/python
 #
 # Urwid canvas class and functions
-#    Copyright (C) 2004-2007  Ian Ward
+#    Copyright (C) 2004-2011  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -19,25 +19,24 @@
 #
 # Urwid web site: http://excess.org/urwid/
 
-from __future__ import generators
 import weakref
 
-from util import *
-from escape import * 
-from text_layout import *
+from urwid.util import rle_len, rle_append_modify, rle_join_modify, rle_product, \
+    calc_width, calc_text_pos, apply_target_encoding, trim_text_attr_cs
+from urwid.text_layout import trim_line, LayoutSegment
+from urwid.compat import bytes
 
-import sys
 
 class CanvasCache(object):
     """
     Cache for rendered canvases.  Automatically populated and
-    accessed by Widget render() MetaClass magic, cleared by 
+    accessed by Widget render() MetaClass magic, cleared by
     Widget._invalidate().
 
     Stores weakrefs to the canvas objects, so an external class
     must maintain a reference for this cache to be effective.
-    At present the Screen classes store the last topmost canvas 
-    after redrawing the screen, keeping the canvases from being 
+    At present the Screen classes store the last topmost canvas
+    after redrawing the screen, keeping the canvases from being
     garbage collected.
 
     _widgets[widget] = {(wcls, size, focus): weakref.ref(canvas), ...}
@@ -58,6 +57,9 @@ class CanvasCache(object):
         wcls -- widget class that contains render() function
         canvas -- rendered canvas with widget_info (widget, size, focus)
         """
+        if not canvas.cacheable:
+            return
+
         assert canvas.widget_info, "Can't store canvas without widget_info"
         widget, size, focus = canvas.widget_info
         def walk_depends(canv):
@@ -65,7 +67,7 @@ class CanvasCache(object):
             Collect all child widgets for determining who we
             depend on.
             """
-            # FIXME: is this recursion necessary?  The cache 
+            # FIXME: is this recursion necessary?  The cache
             # invalidating might work with only one level.
             depends = []
             for x, y, c, pos in canv.children:
@@ -112,7 +114,7 @@ class CanvasCache(object):
             cls.hits += 1 # more stats
         return canv
     fetch = classmethod(fetch)
-    
+
     def invalidate(cls, widget):
         """
         Remove all canvases cached for widget.
@@ -170,7 +172,7 @@ class CanvasCache(object):
     clear = classmethod(clear)
 
 
-        
+
 class CanvasError(Exception):
     pass
 
@@ -178,6 +180,8 @@ class Canvas(object):
     """
     base class for canvases
     """
+    cacheable = True
+
     _finalized_error = CanvasError("This canvas has been finalized. "
         "Use CompositeCanvas to wrap this canvas if "
         "you need to make changes.")
@@ -194,12 +198,12 @@ class Canvas(object):
         value1, value2, value3 -- if not None, raise a helpful error:
             the old Canvas class is now called TextCanvas.
         """
-        if value1 is not None: 
-            raise _renamed_error
+        if value1 is not None:
+            raise self._renamed_error
         self._widget_info = None
         self.coords = {}
         self.shortcuts = {}
-    
+
     def finalize(self, widget, size, focus):
         """
         Mark this canvas as finalized (should not be any future
@@ -222,20 +226,20 @@ class Canvas(object):
 
     def _raise_old_repr_error(self, val=None):
         raise self._old_repr_error
-    
+
     def _text_content(self):
         """
         Return the text content of the canvas as a list of strings,
         one for each row.
         """
-        return ["".join([text for (attr, cs, text) in row])
+        return [bytes().join([text for (attr, cs, text) in row])
             for row in self.content()]
 
     text = property(_text_content, _raise_old_repr_error)
     attr = property(_raise_old_repr_error, _raise_old_repr_error)
     cs = property(_raise_old_repr_error, _raise_old_repr_error)
-    
-    def content(self, trim_left=0, trim_top=0, cols=None, rows=None, 
+
+    def content(self, trim_left=0, trim_top=0, cols=None, rows=None,
             attr=None):
         raise NotImplementedError()
 
@@ -244,7 +248,7 @@ class Canvas(object):
 
     def rows(self):
         raise NotImplementedError()
-    
+
     def content_delta(self):
         raise NotImplementedError()
 
@@ -254,7 +258,7 @@ class Canvas(object):
             return
         return c[:2] # trim off data part
     def set_cursor(self, c):
-        if self.widget_info:
+        if self.widget_info and self.cacheable:
             raise self._finalized_error
         if c is None:
             try:
@@ -265,6 +269,35 @@ class Canvas(object):
         self.coords["cursor"] = c + (None,) # data part
     cursor = property(get_cursor, set_cursor)
 
+    def get_pop_up(self):
+        c = self.coords.get("pop up", None)
+        if not c:
+            return
+        return c
+    def set_pop_up(self, w, left, top, overlay_width, overlay_height):
+        """
+        This method adds pop-up information to the canvas.  This information
+        is intercepted by a PopUpTarget widget higher in the chain to
+        display a pop-up at the given (left, top) position relative to the
+        current canvas.
+
+        :param w: widget to use for the pop-up
+        :type w: widget
+        :param left: x position for left edge of pop-up >= 0
+        :type left: int
+        :param top: y position for top edge of pop-up >= 0
+        :type top: int
+        :param overlay_width: width of overlay in screen columns > 0
+        :type overlay_width: int
+        :param overlay_height: height of overlay in screen rows > 0
+        :type overlay_height: int
+        """
+        if self.widget_info and self.cacheable:
+            raise self._finalized_error
+
+        self.coords["pop up"] = (left, top, (
+            w, overlay_width, overlay_height))
+
     def translate_coords(self, dx, dy):
         """
         Return coords shifted by (dx, dy).
@@ -280,7 +313,7 @@ class TextCanvas(Canvas):
     """
     class for storing rendered text and attributes
     """
-    def __init__(self, text=None, attr=None, cs=None, 
+    def __init__(self, text=None, attr=None, cs=None,
         cursor=None, maxcol=None, check_width=True):
         """
         text -- list of strings, one for each line
@@ -291,17 +324,17 @@ class TextCanvas(Canvas):
         check_width -- check and fix width of all lines in text
         """
         Canvas.__init__(self)
-        if text == None: 
+        if text == None:
             text = []
 
         if check_width:
             widths = []
             for t in text:
-                if type(t) != type(""):
-                    raise CanvasError("Canvas text must be plain strings encoded in the screen's encoding", `text`)
+                if type(t) != bytes:
+                    raise CanvasError("Canvas text must be plain strings encoded in the screen's encoding", repr(text))
                 widths.append( calc_width( t, 0, len(t)) )
         else:
-            assert type(maxcol) == type(0)
+            assert type(maxcol) == int
             widths = [maxcol] * len(text)
 
         if maxcol is None:
@@ -311,30 +344,30 @@ class TextCanvas(Canvas):
             else:
                 maxcol = 0
 
-        if attr == None: 
+        if attr == None:
             attr = [[] for x in range(len(text))]
         if cs == None:
             cs = [[] for x in range(len(text))]
-        
+
         # pad text and attr to maxcol
         for i in range(len(text)):
             w = widths[i]
-            if w > maxcol: 
-                raise CanvasError("Canvas text is wider than the maxcol specified \n%s\n%s\n%s"%(`maxcol`,`widths`,`text`))
+            if w > maxcol:
+                raise CanvasError("Canvas text is wider than the maxcol specified \n%r\n%r\n%r"%(maxcol,widths,text))
             if w < maxcol:
-                text[i] = text[i] + " "*(maxcol-w)
+                text[i] = text[i] + bytes().rjust(maxcol-w)
             a_gap = len(text[i]) - rle_len( attr[i] )
             if a_gap < 0:
-                raise CanvasError("Attribute extends beyond text \n%s\n%s" % (`text[i]`,`attr[i]`) )
+                raise CanvasError("Attribute extends beyond text \n%r\n%r" % (text[i],attr[i]) )
             if a_gap:
                 rle_append_modify( attr[i], (None, a_gap))
-            
+
             cs_gap = len(text[i]) - rle_len( cs[i] )
             if cs_gap < 0:
-                raise CanvasError("Character Set extends beyond text \n%s\n%s" % (`text[i]`,`cs[i]`) )
+                raise CanvasError("Character Set extends beyond text \n%r\n%r" % (text[i],cs[i]) )
             if cs_gap:
                 rle_append_modify( cs[i], (None, cs_gap))
-            
+
         self._attr = attr
         self._cs = cs
         self.cursor = cursor
@@ -343,12 +376,14 @@ class TextCanvas(Canvas):
 
     def rows(self):
         """Return the number of rows in this canvas."""
-        return len(self._text)
+        rows = len(self._text)
+        assert isinstance(rows, int)
+        return rows
 
     def cols(self):
         """Return the screen column width of this canvas."""
         return self._maxcol
-    
+
     def translated_coords(self,dx,dy):
         """
         Return cursor coords shifted by (dx, dy), or None if there
@@ -365,35 +400,35 @@ class TextCanvas(Canvas):
         Return the canvas content as a list of rows where each row
         is a list of (attr, cs, text) tuples.
 
-        trim_left, trim_top, cols, rows may be set by 
+        trim_left, trim_top, cols, rows may be set by
         CompositeCanvas when rendering a partially obscured
         canvas.
         """
         maxcol, maxrow = self.cols(), self.rows()
-        if not cols: 
+        if not cols:
             cols = maxcol - trim_left
         if not rows:
             rows = maxrow - trim_top
-            
+
         assert trim_left >= 0 and trim_left < maxcol
         assert cols > 0 and trim_left + cols <= maxcol
         assert trim_top >=0 and trim_top < maxrow
         assert rows > 0 and trim_top + rows <= maxrow
-        
+
         if trim_top or rows < maxrow:
             text_attr_cs = zip(
                 self._text[trim_top:trim_top+rows],
-                self._attr[trim_top:trim_top+rows], 
+                self._attr[trim_top:trim_top+rows],
                 self._cs[trim_top:trim_top+rows])
         else:
             text_attr_cs = zip(self._text, self._attr, self._cs)
-        
+
         for text, a_row, cs_row in text_attr_cs:
             if trim_left or cols < self._maxcol:
                 text, a_row, cs_row = trim_text_attr_cs(
-                    text, a_row, cs_row, trim_left, 
+                    text, a_row, cs_row, trim_left,
                     trim_left + cols)
-            attr_cs = util.rle_product(a_row, cs_row)
+            attr_cs = rle_product(a_row, cs_row)
             i = 0
             row = []
             for (a, cs), run in attr_cs:
@@ -402,20 +437,20 @@ class TextCanvas(Canvas):
                 row.append((a, cs, text[i:i+run]))
                 i += run
             yield row
-            
+
 
     def content_delta(self, other):
         """
         Return the differences between other and this canvas.
 
-        If other is the same object as self this will return no 
-        differences, otherwise this is the same as calling 
+        If other is the same object as self this will return no
+        differences, otherwise this is the same as calling
         content().
         """
         if other is self:
             return [self.cols()]*self.rows()
         return self.content()
-    
+
 
 
 class BlankCanvas(Canvas):
@@ -433,7 +468,7 @@ class BlankCanvas(Canvas):
         def_attr = None
         if attr and None in attr:
             def_attr = attr[None]
-        line = [(def_attr, None, " "*cols)]
+        line = [(def_attr, None, bytes().rjust(cols))]
         for i in range(rows):
             yield line
 
@@ -442,10 +477,10 @@ class BlankCanvas(Canvas):
 
     def rows(self):
         raise NotImplementedError("BlankCanvas doesn't know its own size!")
-    
+
     def content_delta(self):
         raise NotImplementedError("BlankCanvas doesn't know its own size!")
-        
+
 blank_canvas = BlankCanvas()
 
 
@@ -461,14 +496,14 @@ class SolidCanvas(Canvas):
         self._cs = cs[0][0]
         self.size = cols, rows
         self.cursor = None
-    
+
     def cols(self):
         return self.size[0]
-    
+
     def rows(self):
         return self.size[1]
 
-    def content(self, trim_left=0, trim_top=0, cols=None, rows=None, 
+    def content(self, trim_left=0, trim_top=0, cols=None, rows=None,
             attr=None):
         if cols is None:
             cols = self.size[0]
@@ -489,7 +524,7 @@ class SolidCanvas(Canvas):
         if other is self:
             return [self.cols()]*self.rows()
         return self.content()
-        
+
 
 
 
@@ -503,15 +538,15 @@ class CompositeCanvas(Canvas):
 
         if canv is a CompositeCanvas, make a copy of its contents
         """
-        # a "shard" is a (num_rows, list of cviews) tuple, one for 
+        # a "shard" is a (num_rows, list of cviews) tuple, one for
         # each cview starting in this shard
 
         # a "cview" is a tuple that defines a view of a canvas:
         # (trim_left, trim_top, cols, rows, attr_map, canv)
 
         # a "shard tail" is a list of tuples:
-        # (col_gap, done_rows, content_iter, cview) 
-        
+        # (col_gap, done_rows, content_iter, cview)
+
         # tuples that define the unfinished cviews that are part of
         # shards following the first shard.
         Canvas.__init__(self)
@@ -524,7 +559,7 @@ class CompositeCanvas(Canvas):
                 self.shards = canv.shards
             else:
                 self.shards = [(canv.rows(), [
-                    (0, 0, canv.cols(), canv.rows(), 
+                    (0, 0, canv.cols(), canv.rows(),
                     None, canv)])]
             self.children = [(0, 0, canv, None)]
             self.coords.update(canv.coords)
@@ -532,14 +567,23 @@ class CompositeCanvas(Canvas):
                 self.shortcuts[shortcut] = "wrap"
 
     def rows(self):
-        return sum([r for r,cv in self.shards])
+        for r,cv in self.shards:
+            try:
+                assert isinstance(r, int)
+            except AssertionError:
+                raise AssertionError(r, cv)
+        rows = sum([r for r,cv in self.shards])
+        assert isinstance(rows, int)
+        return rows
 
     def cols(self):
         if not self.shards:
             return 0
-        return sum([cv[2] for cv in self.shards[0][1]])
+        cols = sum([cv[2] for cv in self.shards[0][1]])
+        assert isinstance(cols, int)
+        return cols
+
 
-        
     def content(self):
         """
         Return the canvas content as a list of rows where each row
@@ -554,11 +598,11 @@ class CompositeCanvas(Canvas):
             for i in range(num_rows):
                 yield shard_body_row(sbody)
 
-            # prepare next shard tail            
+            # prepare next shard tail
             shard_tail = shard_body_tail(num_rows, sbody)
 
-                    
-    
+
+
     def content_delta(self, other):
         """
         Return the differences between other and this canvas.
@@ -577,16 +621,16 @@ class CompositeCanvas(Canvas):
             # output rows
             row = []
             for i in range(num_rows):
-                # if whole shard is unchanged, don't keep 
+                # if whole shard is unchanged, don't keep
                 # calling shard_body_row
-                if len(row) != 1 or type(row[0]) != type(0):
+                if len(row) != 1 or type(row[0]) != int:
                     row = shard_body_row(sbody)
                 yield row
 
             # prepare next shard tail
             shard_tail = shard_body_tail(num_rows, sbody)
-                
-    
+
+
     def trim(self, top, count=None):
         """Trim lines from the top and/or bottom of canvas.
 
@@ -598,7 +642,7 @@ class CompositeCanvas(Canvas):
             top, self.rows())
         if self.widget_info:
             raise self._finalized_error
-        
+
         if top:
             self.shards = shards_trim_top(self.shards, top)
 
@@ -609,10 +653,10 @@ class CompositeCanvas(Canvas):
 
         self.coords = self.translate_coords(0, -top)
 
-        
+
     def trim_end(self, end):
         """Trim lines from the bottom of the canvas.
-        
+
         end -- number of lines to remove from the end
         """
         assert end > 0, "invalid trim amount %d!"%end
@@ -620,14 +664,14 @@ class CompositeCanvas(Canvas):
             end, self.rows())
         if self.widget_info:
             raise self._finalized_error
-        
+
         self.shards = shards_trim_rows(self.shards, self.rows() - end)
 
-            
+
     def pad_trim_left_right(self, left, right):
         """
         Pad or trim this canvas on the left and right
-        
+
         values > 0 indicate screen columns to pad
         values < 0 indicate screen columns to trim
         """
@@ -677,26 +721,26 @@ class CompositeCanvas(Canvas):
                 [(0,0,cols,top,None,blank_canvas)])] + \
                 self.shards
             self.coords = self.translate_coords(0, top)
-        
+
         if bottom > 0:
             if orig_shards is self.shards:
                 self.shards = self.shards[:]
             self.shards.append((bottom,
                 [(0,0,cols,bottom,None,blank_canvas)]))
 
-        
+
     def overlay(self, other, left, top ):
         """Overlay other onto this canvas."""
         if self.widget_info:
             raise self._finalized_error
-        
+
         width = other.cols()
         height = other.rows()
         right = self.cols() - left - width
         bottom = self.rows() - top - height
-        
-        assert right >= 0, "top canvas of overlay not the size expected!" + `other.cols(),left,right,width`
-        assert bottom >= 0, "top canvas of overlay not the size expected!" + `other.rows(),top,bottom,height`
+
+        assert right >= 0, "top canvas of overlay not the size expected!" + repr((other.cols(),left,right,width))
+        assert bottom >= 0, "top canvas of overlay not the size expected!" + repr((other.rows(),top,bottom,height))
 
         shards = self.shards
         top_shards = []
@@ -711,22 +755,22 @@ class CompositeCanvas(Canvas):
 
         left_shards = []
         right_shards = []
-        if left:
+        if left > 0:
             left_shards = [shards_trim_sides(side_shards, 0, left)]
-        if right:
-            right_shards = [shards_trim_sides(side_shards, 
-                left + width, right)]
-        
+        if right > 0:
+            right_shards = [shards_trim_sides(side_shards,
+                max(0, left + width), right)]
+
         if not self.rows():
             middle_shards = []
         elif left or right:
-            middle_shards = shards_join(left_shards + 
+            middle_shards = shards_join(left_shards +
                 [other.shards] + right_shards)
         else:
             middle_shards = other.shards
 
         self.shards = top_shards + middle_shards + bottom_shards
-        
+
         self.coords.update(other.translate_coords(left, top))
 
 
@@ -736,7 +780,7 @@ class CompositeCanvas(Canvas):
         attribute currently set to None, leaving other attributes
         intact."""
         self.fill_attr_apply({None:a})
-    
+
     def fill_attr_apply(self, mapping):
         """
         Apply an attribute-mapping dictionary to the canvas.
@@ -752,7 +796,7 @@ class CompositeCanvas(Canvas):
             for cv in original_cviews:
                 # cv[4] == attr_map
                 if cv[4] is None:
-                    new_cviews.append(cv[:4] + 
+                    new_cviews.append(cv[:4] +
                         (mapping,) + cv[5:])
                 else:
                     combined = dict(mapping)
@@ -787,7 +831,7 @@ def shard_body_row(sbody):
             row.extend(content_iter.next())
         else:
             # need to skip this unchanged canvas
-            if row and type(row[-1]) == type(0):
+            if row and type(row[-1]) == int:
                 row[-1] = row[-1] + cview[2]
             else:
                 row.append(cview[2])
@@ -815,7 +859,7 @@ def shard_body_tail(num_rows, sbody):
 
 def shards_delta(shards, other_shards):
     """
-    Yield shards1 with cviews that are the same as shards2 
+    Yield shards1 with cviews that are the same as shards2
     having canv = None.
     """
     other_shards_iter = iter(other_shards)
@@ -866,11 +910,11 @@ def shard_cviews_delta(cviews, other_cviews):
 
 def shard_body(cviews, shard_tail, create_iter=True, iter_default=None):
     """
-    Return a list of (done_rows, content_iter, cview) tuples for 
+    Return a list of (done_rows, content_iter, cview) tuples for
     this shard and shard tail.
 
-    If a canvas in cviews is None (eg. when unchanged from 
-    shard_cviews_delta()) or if create_iter is False then no 
+    If a canvas in cviews is None (eg. when unchanged from
+    shard_cviews_delta()) or if create_iter is False then no
     iterator is created for content_iter.
 
     iter_default is the value used for content_iter when no iterator
@@ -894,7 +938,7 @@ def shard_body(cviews, shard_tail, create_iter=True, iter_default=None):
                 raise CanvasError("cviews overflow gaps in"
                     " shard_tail!")
             if create_iter and canv:
-                new_iter = canv.content(trim_left, trim_top, 
+                new_iter = canv.content(trim_left, trim_top,
                     cols, rows, attr_map)
             else:
                 new_iter = iter_default
@@ -904,7 +948,7 @@ def shard_body(cviews, shard_tail, create_iter=True, iter_default=None):
         (trim_left, trim_top, cols, rows, attr_map, canv) = \
             cview[:6]
         if create_iter and canv:
-            new_iter = canv.content(trim_left, trim_top, cols, rows, 
+            new_iter = canv.content(trim_left, trim_top, cols, rows,
                 attr_map)
         else:
             new_iter = iter_default
@@ -928,20 +972,20 @@ def shards_trim_top(shards, top):
         shard_tail = shard_body_tail(num_rows, sbody)
         top -= num_rows
     else:
-        raise CanvasError("tried to trim shards out of existance")
-    
+        raise CanvasError("tried to trim shards out of existence")
+
     sbody = shard_body(cviews, shard_tail, False)
     shard_tail = shard_body_tail(num_rows, sbody)
     # trim the top of this shard
     new_sbody = []
     for done_rows, content_iter, cv in sbody:
-        new_sbody.append((0, content_iter, 
+        new_sbody.append((0, content_iter,
             cview_trim_top(cv, done_rows+top)))
     sbody = new_sbody
-    
-    new_shards = [(num_rows-top, 
+
+    new_shards = [(num_rows-top,
         [cv for done_rows, content_iter, cv in sbody])]
-    
+
     # write out the rest of the shards
     new_shards.extend(shard_iter)
 
@@ -953,7 +997,6 @@ def shards_trim_rows(shards, keep_rows):
     """
     assert keep_rows >= 0, keep_rows
 
-    shard_tail = []
     new_shards = []
     done_rows = 0
     for num_rows, cviews in shards:
@@ -962,7 +1005,7 @@ def shards_trim_rows(shards, keep_rows):
         new_cviews = []
         for cv in cviews:
             if cv[3] + done_rows > keep_rows:
-                new_cviews.append(cview_trim_rows(cv, 
+                new_cviews.append(cview_trim_rows(cv,
                     keep_rows - done_rows))
             else:
                 new_cviews.append(cv)
@@ -979,7 +1022,7 @@ def shards_trim_sides(shards, left, cols):
     """
     Return shards with starting from column left and cols total width.
     """
-    assert left >= 0 and cols > 0
+    assert left >= 0 and cols > 0, (left, cols)
     shard_tail = []
     new_shards = []
     right = left + cols
@@ -1043,7 +1086,7 @@ def shards_join(shard_lists):
 
 def cview_trim_rows(cv, rows):
     return cv[:3] + (rows,) + cv[4:]
-    
+
 def cview_trim_top(cv, trim):
     return (cv[0], trim + cv[1], cv[2], cv[3] - trim) + cv[4:]
 
@@ -1054,15 +1097,19 @@ def cview_trim_cols(cv, cols):
     return cv[:2] + (cols,) + cv[3:]
 
 
-        
+
 
 def CanvasCombine(l):
     """Stack canvases in l vertically and return resulting canvas.
 
-    l -- list of (canvas, position, focus) tuples.  position is a value
-         that widget.set_focus will accept, or None if not allowed.
-         focus is True if this canvas is the one that would be in focus
-         if the whole widget is in focus.
+    :param l: list of (canvas, position, focus) tuples:
+
+              position
+                a value that widget.set_focus will accept or None if not
+                allowed
+              focus
+                True if this canvas is the one that would be in focus
+                if the whole widget is in focus
     """
     clist = [(CompositeCanvas(c),p,f) for c,p,f in l]
 
@@ -1073,7 +1120,7 @@ def CanvasCombine(l):
     focus_index = 0
     n = 0
     for canv, pos, focus in clist:
-        if focus: 
+        if focus:
             focus_index = n
         children.append((0, row, canv, pos))
         shards.extend(canv.shards)
@@ -1082,7 +1129,7 @@ def CanvasCombine(l):
             combined_canvas.shortcuts[shortcut] = pos
         row += canv.rows()
         n += 1
-    
+
     if focus_index:
         children = [children[focus_index]] + children[:focus_index] + \
             children[focus_index+1:]
@@ -1098,7 +1145,7 @@ def CanvasOverlay(top_c, bottom_c, left, top):
     """
     overlayed_canvas = CompositeCanvas(bottom_c)
     overlayed_canvas.overlay(top_c, left, top)
-    overlayed_canvas.children = [(left, top, top_c, None), 
+    overlayed_canvas.children = [(left, top, top_c, None),
         (0, 0, bottom_c, None)]
     overlayed_canvas.shortcuts = {} # disable background shortcuts
     for shortcut in top_c.shortcuts.keys():
@@ -1109,18 +1156,24 @@ def CanvasOverlay(top_c, bottom_c, left, top):
 def CanvasJoin(l):
     """
     Join canvases in l horizontally. Return result.
-    l -- list of (canvas, position, focus, cols) tuples.  position is a 
-         value that widget.set_focus will accept,  or None if not allowed.
-         focus is True if this canvas is the one that would be in focus if
-         the whole widget is in focus.  cols is the number of screen
-         columns that this widget will require, if larger than the actual
-         canvas.cols() value then this widget will be padded on the right.
+
+    :param l: list of (canvas, position, focus, cols) tuples:
+
+              position
+                value that widget.set_focus will accept or None if not allowed
+              focus
+                True if this canvas is the one that would be in focus if
+                the whole widget is in focus
+              cols
+                is the number of screen columns that this widget will require,
+                if larger than the actual canvas.cols() value then this widget
+                will be padded on the right.
     """
-    
+
     l2 = []
     focus_item = 0
     maxrow = 0
-    n = 0 
+    n = 0
     for canv, pos, focus, cols in l:
         rows = canv.rows()
         pad_right = cols - canv.cols()
@@ -1130,7 +1183,7 @@ def CanvasJoin(l):
             maxrow = rows
         l2.append((canv, pos, pad_right, rows))
         n += 1
-    
+
     shard_lists = []
     children = []
     joined_canvas = CompositeCanvas()
@@ -1158,17 +1211,16 @@ def CanvasJoin(l):
 
 
 def apply_text_layout(text, attr, ls, maxcol):
-    utext = type(text)==type(u"")
     t = []
     a = []
     c = []
-    
+
     class AttrWalk:
         pass
     aw = AttrWalk
     aw.k = 0 # counter for moving through elements of a
     aw.off = 0 # current offset into text of attr[ak]
-    
+
     def arange( start_offs, end_offs ):
         """Return an attribute list for the range of text specified."""
         if start_offs < aw.off:
@@ -1194,19 +1246,19 @@ def apply_text_layout(text, attr, ls, maxcol):
             aw.off += run
         return o
 
-    
+
     for line_layout in ls:
         # trim the line to fit within maxcol
         line_layout = trim_line( line_layout, text, 0, maxcol )
-        
+
         line = []
         linea = []
         linec = []
-            
+
         def attrrange( start_offs, end_offs, destw ):
             """
             Add attributes based on attributes between
-            start_offs and end_offs. 
+            start_offs and end_offs.
             """
             if start_offs == end_offs:
                 [(at,run)] = arange(start_offs,end_offs)
@@ -1225,12 +1277,12 @@ def apply_text_layout(text, attr, ls, maxcol):
                 tseg = text[o:o+run]
                 tseg, cs = apply_target_encoding( tseg )
                 segw = rle_len(cs)
-                
+
                 rle_append_modify( linea, ( at, segw ))
                 o += run
                 destw -= segw
-            
-            
+
+
         for seg in line_layout:
             #if seg is None: assert 0, ls
             s = LayoutSegment(seg)
@@ -1247,17 +1299,17 @@ def apply_text_layout(text, attr, ls, maxcol):
                 rle_join_modify( linec, cs )
             elif s.offs:
                 if s.sc:
-                    line.append(" "*s.sc)
+                    line.append(bytes().rjust(s.sc))
                     attrrange( s.offs, s.offs, s.sc )
             else:
-                line.append(" "*s.sc)
+                line.append(bytes().rjust(s.sc))
                 linea.append((None, s.sc))
                 linec.append((None, s.sc))
-            
-        t.append("".join(line))
+
+        t.append(bytes().join(line))
         a.append(linea)
         c.append(linec)
-        
+
     return TextCanvas(t, a, c, maxcol=maxcol)
 
 
